from abc import ABC, abstractmethod

from typing import List


class MetricNamer(ABC):
    @abstractmethod
    def __call__(self, scope) -> str:
        pass  # pragma: no cover


class TimingClient(ABC):
    """ An abstract class detailing the client interface """

    @abstractmethod
    def timing(self, metric_name: str, timing: float, tags: List[str]):
        pass  # pragma: no cover



from starlette.routing import Match

from timing_asgi.interfaces import MetricNamer
from timing_asgi.utils import PathToName


class StarletteScopeToName(MetricNamer):
    """Class which extracts a suitable metric name from a matching Starlette route.

    Uses the route name if it has one, but otherwise constructs a name based on the
    full module + method/class path.

    Falls back to either a provided fallback callable, or uses PathToName, which is the
    TimingMiddleware default.
    """

    def __init__(self, prefix, starlette_app, fallback=None):
        self.prefix = prefix
        self.starlette_app = starlette_app
        if fallback is None:
            fallback = PathToName(prefix)
        self.fallback = fallback

    def __call__(self, scope):
        route = None
        for r in self.starlette_app.router.routes:
            if r.matches(scope)[0] == Match.FULL:
                route = r
                break
        if route is not None:
            return f"{self.prefix}.{route.endpoint.__module__}.{route.name}"
        else:
            return self.fallback(scope)



# Use timer that's not susceptible to time of day adjustments.
try:
    # perf_counter is only present on Py3.3+
    from time import perf_counter as time_now
except ImportError:
    # fall back to using time
    from time import time as time_now

from .interfaces import MetricNamer


class TimingStats(object):
    def __init__(self, name=None):
        self.name = name

    def start(self):
        self.start_time = time_now()

    def stop(self):
        self.end_time = time_now()

    def __enter__(self):
        self.start()
        return self

    @property
    def time(self):
        return self.end_time - self.start_time

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()


class PathToName(MetricNamer):
    def __init__(self, prefix):
        self.prefix = prefix

    def __call__(self, scope):
        path = scope.get("path")[1:]
        return f"{self.prefix}.{path.replace('/', '.')}"
